#ifndef DUNE_MICROSTRUCTURE_PRESTRAINEDMATERIAL_HH
#define DUNE_MICROSTRUCTURE_PRESTRAINEDMATERIAL_HH

#include <dune/grid/uggrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>
#include <dune/functions/gridfunctions/gridviewentityset.hh>
#include <dune/common/parametertree.hh>
#include <dune/fufem/dunepython.hh>
#include <dune/microstructure/matrix_operations.hh>
#include <dune/microstructure/voigthelper.hh>

// #include <dune/microstructure/materialDefinitions.hh> //deprecated


using namespace MatrixOperations;
using namespace Dune::Functions;
using namespace Dune::Functions::BasisFactory;
using std::pow;
using std::abs;
using std::sqrt;
using std::sin;
using std::cos;

using std::shared_ptr;
using std::make_shared;






//   // ----------------------------------------------------------------------------------
//   template<class Domain>
//   int indicatorFunction_material(const Domain& x)
//   {
//     double theta=0.25;
//     if (x[0] <(-0.5+theta) and x[2]<(-0.5+theta))
//         return 1;    //#Phase1
//     else if (x[1]<(-0.5+theta) and x[2]>(0.5-theta))
//         return 2;    //#Phase2
//     else 
//         return 3;    //#Phase3
//   }
//  MatrixRT getL(const int& phase)
//   {
//       const FieldVector<double,3> mu     = {80.0, 80.0, 60.0};
//       const FieldVector<double,3> lambda = {80.0, 80.0, 25.0};



//       if (phase == 1)
//         return 2.0 * mu[0] * sym(G) + lambda[0] * trace(sym(G)) * Id(); //#Phase1 //Isotrop(mu,lambda)
//       else if (phase == 2)
//         return 2.0 * mu[1] * sym(G) + lambda[1] * trace(sym(G)) * Id(); //#Phase2
//       else 
//         return 2.0 * mu[2] * sym(G) + lambda[2] * trace(sym(G)) * Id(); //#Phase3
//   }
//   MatrixRT prestrain_material(const int& phase)
//   {
//       if (phase == 1)
//         return {{1.0, 0.0, 0.0},    //#Phase1
//                 {0.0, 1.0, 0.0},
//                 {0.0, 0.0, 1.0}
//                };
//       else if (phase == 2)
//         return {{1.0, 0.0, 0.0},    //#Phase2
//                 {0.0, 1.0, 0.0},
//                 {0.0, 0.0, 1.0}
//                };
//       else 
//         return {{0.0, 0.0, 0.0},    //#Phase3
//                 {0.0, 0.0, 0.0},
//                 {0.0, 0.0, 0.0}
//                };
//   }


//--------------------------------------------------------
template <class GridView>         
class prestrainedMaterial
{

public:
  static const int dimworld = 3;  //GridView::dimensionworld;
  static const int dim = 3;       //Domain::dimension;
  using Domain = typename GridView::template Codim<0>::Geometry::GlobalCoordinate;
  using LocalDomain = typename GridView::template Codim<0>::Geometry::LocalCoordinate;
  using Element = typename GridViewEntitySet<GridView, 0>::Element;
  using ScalarRT = Dune::FieldVector< double, 1>;
  using VectorRT = Dune::FieldVector< double, dimworld>;
  using MatrixRT = Dune::FieldMatrix< double, dimworld, dimworld>;
  using FuncScalar = std::function< double(const Domain&) >;
  using Func2int = std::function< int(const Domain&) >;
  using Func2Tensor = std::function< MatrixRT(const Domain&) >;
  using Func2TensorParam = std::function< MatrixRT(const MatrixRT& ,const Domain&) >;
  using Func2TensorPhase = std::function< MatrixRT(const MatrixRT& ,const int&) >;
  using MatrixFunc  = std::function< MatrixRT(const MatrixRT&) >;
  using MatrixPhase = std::function< MatrixRT(const int&) >;

protected:
  const GridView& gridView_;
  const Dune::ParameterTree parameterSet_;
  // std::string materialFunctionName_;

  // Elasticity tensors (in Voigt notation)
  std::vector<VoigtMatrix<double,3> > L_;

  std::vector<Func2Tensor> prestrain_;

  Python::Module module_;

  Func2TensorPhase elasticityTensor_;


  Func2int indicatorFunction_;
  GridViewFunction<int(const Domain&), GridView> indicatorFunctionGVF_;
  LocalFunction<int(const Domain&), typename GridViewEntitySet<GridView, 0>::Element >  localIndicatorFunction_;


  //Transformation matrix for Voigt notation
  Dune::FieldMatrix<double,6,6> D_ = {{1.0, 0.0, 0.0, 0.0 , 0.0, 0.0},
                                      {0.0, 1.0, 0.0, 0.0 , 0.0, 0.0},
                                      {0.0, 0.0, 1.0, 0.0 , 0.0, 0.0},
                                      {0.0, 0.0, 0.0, sqrt(2.0) , 0.0, 0.0},
                                      {0.0, 0.0, 0.0, 0.0 , sqrt(2.0), 0.0},
                                      {0.0, 0.0, 0.0, 0.0 , 0.0, sqrt(2.0)}
                                     };
  Dune::FieldMatrix<double,6,6> D_inv = {{1.0, 0.0, 0.0, 0.0 , 0.0, 0.0},
                                         {0.0, 1.0, 0.0, 0.0 , 0.0, 0.0},
                                         {0.0, 0.0, 1.0, 0.0 , 0.0, 0.0},
                                         {0.0, 0.0, 0.0, 1.0/sqrt(2.0) , 0.0, 0.0},
                                         {0.0, 0.0, 0.0, 0.0 , 1.0/sqrt(2.0), 0.0},
                                         {0.0, 0.0, 0.0, 0.0 , 0.0, 1.0/sqrt(2.0)}
                                        };



public: 
    ///////////////////////////////
    // constructor
    ///////////////////////////////
    // prestrainedMaterial(const GridView gridView,
    //                     const Dune::ParameterTree& parameterSet)
    //                    : gridView_(gridView), 
    //                      parameterSet_(parameterSet)
    // {
    prestrainedMaterial(const GridView gridView,
                        const Dune::ParameterTree& parameterSet,
                        Python::Module pyModule)
                       : gridView_(gridView), 
                         parameterSet_(parameterSet),
                         module_(pyModule) 
    {
       setup();
    } 

/**
 * @brief TODO: Code allows a spatially varying prestrain. However most of the time we only 
 *              use a constant prestrain. 
 *              - Add switch (constant/non-constant) prestrain?
 * 
 * @param phase           material phase
 * @return Func2Tensor 
 */
Func2Tensor setupPrestrainPhase(const int phase){

  auto prestrain = Python::make_function<MatrixRT>(module_.get("prestrain_phase" + std::to_string(phase)));

  int axis = 0;
  double angle = 0;
  try 
  {
    // module_.get("phase" + std::to_string(phase) + "_axis").toC<int>(axis);
    // module_.get("phase" + std::to_string(phase) + "_angle").toC<double>(angle);
    axis = parameterSet_.get<int>("phase" + std::to_string(phase) + "_axis", 0);
    angle = parameterSet_.get<double>("phase" + std::to_string(phase) + "_angle", 0);
  }
  catch(Dune::Exception&)
  {
    //default frame is used.
  }
  /**
   * @brief (optional) Tramsform prestrain to the correct frame.
   */
  if(abs(angle) > 1e-12){
    auto R = rotationMatrix(axis,angle);
    Func2Tensor f = [prestrain,R] (const Domain& x) { return (prestrain(x).leftmultiply(R)).rightmultiply(R.transposed()); };

    return f;
  }
  else 
    return prestrain;
}



// Dune::FieldMatrix<double,6,6> setupPhase(const std::string phaseType, Python::Module module, int phase)
Dune::FieldMatrix<double,6,6> setupPhase(const int phase)
{
  std::string phaseType;
  // module_.get("phase" + std::to_string(phase) + "_type").toC<std::string>(phaseType);
  phaseType = parameterSet_.get<std::string>("phase" + std::to_string(phase) + "_type");

  std::cout << "Setup phase "<< phase << " as " << phaseType << " material." << std::endl;

  if(phaseType == "isotropic")
  {
      Dune::FieldVector<double,2> materialParameters(0);
      module_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldVector<double,2>>(materialParameters);
      std::cout << "Lame-Parameters (mu,lambda): (" << materialParameters[0] << "," << materialParameters[1] << ") " << std::endl;
      return isotropic(materialParameters[0],materialParameters[1]);
  }

  /**
   * @brief For other material classes. The stiffness-/compliance matrix depends on the 
   *        coordinate frame.
   * 
   *   The default frame vectors:
   */
  // VectorRT frameVector1 = {1.0,0.0,0.0};
  // VectorRT frameVector2 = {0.0,1.0,0.0};
  // VectorRT frameVector3 = {0.0,0.0,1.0};
  /**
   * @brief For other material classes. The stiffness-/compliance matrix depends on the 
   *        coordinate frame.
   * 
   *        (optional) rotate material coordinate frame by an 'angle' around 'axis'
   */
  int axis = 0;
  double angle = 0;

  try 
  {
    // module.get("phase" + std::to_string(phase) + "_FrameVector1").toC<VectorRT>(frameVector1);
    // module.get("phase" + std::to_string(phase) + "_FrameVector2").toC<VectorRT>(frameVector2);
    // module.get("phase" + std::to_string(phase) + "_FrameVector3").toC<VectorRT>(frameVector3);
    // printvector(std::cout, frameVector1, "frameVector1: ", "--");
    // printvector(std::cout, frameVector2, "frameVector2: ", "--");
    // printvector(std::cout, frameVector3, "frameVector3: ", "--");
    // module_.get("phase" + std::to_string(phase) + "_axis").toC<int>(axis);
    // module_.get("phase" + std::to_string(phase) + "_angle").toC<double>(angle);
    axis = parameterSet_.get<int>("phase" + std::to_string(phase) + "_axis", 0);
    angle = parameterSet_.get<double>("phase" + std::to_string(phase) + "_angle", 0);
  }
  catch(Dune::Exception& e)
  {
    // std::cout << "(Could not read frame vectors for material phase) " << phase << ": - Canonical Frame (default) is used." << std::endl;
    std::cout << "(Could not read rotation axis & angle for material phase) " << phase << ": - Canonical Frame (default) is used." << std::endl;
  }


  if(phaseType == "orthotropic")
  {
      Dune::FieldVector<double,9> materialParameters(0);
      module_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldVector<double,9>>(materialParameters);
      // return orthotropic(frameVector1,frameVector2,frameVector3,materialParameters);
      return orthotropic(axis,angle,materialParameters);
  }
  if(phaseType == "transversely_isotropic")
  {
      Dune::FieldVector<double,5> materialParameters(0);
      module_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldVector<double,5>>(materialParameters);
      // return transversely_isotropic(frameVector1,frameVector2,frameVector3,materialParameters);
      return transversely_isotropic(axis,angle,materialParameters);
  }
  if(phaseType == "general_anisotropic")
  {
      Dune::FieldMatrix<double,6,6> materialParameters(0); //compliance matric
      module_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldMatrix<double,6,6>>(materialParameters);
      // return general_anisotropic(frameVector1,frameVector2,frameVector3,materialParameters);
      return general_anisotropic(axis,angle,materialParameters);
  }
  else
    DUNE_THROW(Dune::Exception, " Unknown material class for phase " + std::to_string(phase));
}



  /**
   * @brief Setup method that determines the correct stiffness-matrix (constitutive law)
   *        for each material phase separately. 
   * 
   */
  void setup()
  {
    indicatorFunction_ = Python::make_function<int>(module_.get("indicatorFunction"));
    indicatorFunctionGVF_ = Dune::Functions::makeGridViewFunction(indicatorFunction_, gridView_);
    localIndicatorFunction_ = localFunction(indicatorFunctionGVF_);

    int phases;
    // module_.get("Phases").toC<int>(phases);
    phases = parameterSet_.get<int>("Phases"); 
    std::cout << "---- Starting material setup... (Number of Phases: "<< phases << ") "  << std::endl;

    L_.resize(phases);
    prestrain_.resize(phases);

    for (int i=0; i<phases; i++)
    {
      L_[i] = setupPhase(i+1);
      prestrain_[i] = setupPrestrainPhase(i+1);
    }

    std::cout << "Material setup done." << std::endl;
  }

  // void setup(const std::string name)  // can't use materialFunctionName_ here?
  // {
  //   Python::Module module = Python::import(name);
  //   indicatorFunction_ = Python::make_function<int>(module.get("indicatorFunction"));

  //   int phases;
  //   module.get("Phases").toC<int>(phases);
  //   std::cout << "---- Starting material setup... (Number of Phases: "<< phases << ") "  << std::endl;

  //   L_.resize(phases);
  //   prestrain_.resize(phases);

  //   for (int i=0; i<phases; i++)
  //   {
  //     L_[i] = setupPhase(module,i+1);
  //     prestrain_[i] = setupPrestrainPhase(module, i+1);
  //   }

  //   std::cout << "Material setup done." << std::endl;
  // }


  ////////////////////////////////////////////////////////
  //      MATERIAL CLASSES DEFINITIONS:
  ////////////////////////////////////////////////////////

  /** \brief Isotropic stiffness matrix in Voigt notation but scaled by a factor of 2 in the last three columns. */
  Dune::FieldMatrix<double,6,6> isotropic(const double& mu, const double& lambda)
  {
        return {{lambda+2.0*mu, lambda       , lambda       , 0.0   , 0.0   , 0.0    },
                {lambda       , lambda+2.0*mu, lambda       , 0.0   , 0.0   , 0.0    },
                {lambda       , lambda       , lambda+2.0*mu, 0.0   , 0.0   , 0.0    },
                {  0.0        ,  0.0         , 0.0          , 2.0*mu, 0.0   , 0.0    },
                {  0.0        ,  0.0         , 0.0          , 0.0   , 2.0*mu, 0.0    },
                {  0.0        ,  0.0         , 0.0          , 0.0   , 0.0   , 2.0*mu }
              };
  }
  //-------------------------------------------------------------------------------

  //--- Definition of orthotropic elasticity tensor (in voigt notation)
  // - Input: (Youngs modulus, shear modulus, Poisson ratio): {E1,E2,E3,G12,G23,G31,nu12,nu13,nu23}
  // - Output: compliance Matrix
  // Dune::FieldMatrix<double,6,6> orthotropic(const VectorRT& framevector1,
  //                                           const VectorRT& framevector2,
  //                                           const VectorRT& framevector3,
  //                                           const Dune::FieldVector<double,9>& p //elastic moduli {E1,E2,E3,G12,G23,G31,nu12,nu13,nu23}
  //                                           ) 

  /**
   * @brief Definition of orthotropic elasticity/stiffness tensor (in Voigt notation).
   *        The convention in literature is to provide the elastic constants in form of the 'compliance matrix'
   *        In order to compute stresses from strain, we have to invert the 'compliance matrix' to obtain the 'stiffness matrix'
   *        (see https://en.wikipedia.org/wiki/Orthotropic_material for more details)
   * 
   * @param axis    rotate material frame around axis
   * @param angle   rotation angle 
   * @param p       elastic constants: (E1,E2,E3,G12,G23,G31,nu12,nu13,nu23)
   * @return        Stiffness matrix
   */
  Dune::FieldMatrix<double,6,6> orthotropic(const int& axis,
                                            const double& angle,
                                            const Dune::FieldVector<double,9>& p
                                            ) 
  {
      auto E_1 = p[0];
      auto E_2 = p[1];
      auto E_3 = p[2];
      auto G_12 = p[3];
      auto G_23 = p[4];
      auto G_31 = p[5];
      auto nu_12 = p[6];
      auto nu_13 = p[7];
      auto nu_23 = p[8];
      //other three poisson ratios are not independent 
      auto nu_21 = (p[6]/p[0])*p[1];
      auto nu_31 = (p[7]/p[0])*p[2];
      auto nu_32 = (p[8]/p[1])*p[2]; 

      std::string axis_string;
      switch (axis)
      {
        case 0: axis_string = "x - axis";
        case 1: axis_string = "y - axis";
        case 2: axis_string = "z - axis";
        default: break;
      }

      // Compliance matrix
      Dune::FieldMatrix<double,6,6> Compliance = {{1.0/E_1      , -nu_21/E_2   , -nu_31/E_3   , 0.0     , 0.0     , 0.0     },
                                                  {-nu_12/E_1   , 1.0/E_2      , -nu_32/E_3   , 0.0     , 0.0     , 0.0     },
                                                  {-nu_13/E_1   , -nu_23/E_2   , 1.0/E_3      , 0.0     , 0.0     , 0.0     },
                                                  {  0.0        ,  0.0         , 0.0          , 1.0/G_23, 0.0     , 0.0     },
                                                  {  0.0        ,  0.0         , 0.0          , 0.0     , 1.0/G_31, 0.0     },
                                                  {  0.0        ,  0.0         , 0.0          , 0.0     , 0.0     , 1.0/G_12}
                                                  };


      /**
       * @brief (optional) Tramsform compliance matrix to the correct frame.
       */
      if(abs(angle) > 1e-12){
        // std::cout << "material frame rotated around axis: " << axis << " by an angle of: " << angle << std::endl;
        std::cout << "material frame rotated around axis: " << axis_string << " by an angle of: " << angle << std::endl;
        Dune::FieldMatrix<double,6,6> C_rot = rotationMatrixCompliance(axis,angle);
        (Compliance.leftmultiply(C_rot)).rightmultiply(C_rot.transposed());
      }

      printmatrix(std::cout, Compliance, "Compliance matrix used:", "--");

      // invert Compliance matrix to obtain elasticity/stiffness matrix
      Compliance.invert();
      // printmatrix(std::cout, Compliance, "Stiffness Tensor", "--");
      scaleStiffnessMatrix(Compliance);
      // printmatrix(std::cout, Compliance, "Stiffness Tensor scaled", "--");
      // printmatrix(std::cout, Compliance, "Stiffness Tensor", "--");
      return Compliance;
  }
  //-------------------------------------------------------------------------------


  // Dune::FieldMatrix<double,6,6> transversely_isotropic(const VectorRT& framevector1,
  //                                                      const VectorRT& framevector2,
  //                                                      const VectorRT& framevector3,
  //                                                      const Dune::FieldVector<double,5>& p //elastic moduli {E1,E2,G12,nu12,nu23}
  //                                                      ) 
  //--- Definition of transversely isotropic elasticity tensor (in voigt notation)
  // - Input: (Youngs modulus, shear modulus, Poisson ratio): {E1,E2,G12,nu12,nu23}
  // - Output: compliance Matrix
  Dune::FieldMatrix<double,6,6> transversely_isotropic(const int& axis,
                                          const double& angle,
                                          const Dune::FieldVector<double,5>& p
                                          ) 
  {
    // (see https://en.wikipedia.org/wiki/Poisson%27s_ratio)
      auto E_1 = p[0];
      auto E_2 = p[1];
      auto E_3 = E_2;
      auto nu_12 = p[3];
      auto nu_13 = nu_12;
      auto nu_21 = (nu_12/E_1)*E_2;
      auto nu_31 = nu_21;
      auto nu_23 = p[4];
      auto nu_32 = nu_23;
      auto G_12 = p[2];
      auto G_23 = E_2/(2.0*(1+nu_23));
      auto G_31 = G_23;
      //other three poisson ratios are not independent 

      std::string axis_string;
      switch (axis)
      {
        case 0: axis_string = "x - axis";
        case 1: axis_string = "y - axis";
        case 2: axis_string = "z - axis";
        default: break;
      }
      
      //---compliance matrix
      Dune::FieldMatrix<double,6,6> Compliance = {{1.0/E_1      , -nu_21/E_2   , -nu_31/E_3   , 0.0     , 0.0     , 0.0   },
                                                  {-nu_12/E_1   , 1.0/E_2      , -nu_32/E_3   , 0.0     , 0.0     , 0.0   },
                                                  {-nu_13/E_1   , -nu_23/E_2   , 1.0/E_3      , 0.0     , 0.0     , 0.0   },
                                                  {  0.0        ,  0.0         , 0.0          , 1.0/G_23, 0.0     , 0.0   },
                                                  {  0.0        ,  0.0         , 0.0          , 0.0     , 1.0/G_31, 0.0   },
                                                  {  0.0        ,  0.0         , 0.0          , 0.0     , 0.0     , 1.0/G_12}
                                                  };

      /**
       * @brief (optional) Tramsform compliance matrix to the correct frame.
       */
      if(abs(angle) > 1e-12){
        // std::cout << "material frame rotated around axis: " << axis << " by an angle of: " << angle << std::endl;
        std::cout << "material frame rotated around axis: " << axis_string << " by an angle of: " << angle << std::endl;
        Dune::FieldMatrix<double,6,6> C_rot = rotationMatrixCompliance(axis,angle);
        (Compliance.leftmultiply(C_rot)).rightmultiply(C_rot.transposed());
      }

      printmatrix(std::cout, Compliance, "Compliance matrix used:", "--");
      // printmatrix(std::cout, C, "C", "--");

      //--- return elasticity tensor 
      Compliance.invert();
      // printmatrix(std::cout, Compliance, "Stiffness Tensor", "--");
      
      scaleStiffnessMatrix(Compliance);

      return Compliance;
  }
  //-------------------------------------------------------------------------------


  // --- Definition of transversely isotropic elasticity tensor (in voigt notation)
  // - Input: (Youngs modulus, shear modulus, Poisson ratio): {E1,E2,G12,nu12,nu23}
  // - Output: inverse compliance Matrix
  // Dune::FieldMatrix<double,6,6> general_anisotropic(const VectorRT& framevector1,
  //                                                   const VectorRT& framevector2,
  //                                                   const VectorRT& framevector3,
  //                                                   const Dune::FieldMatrix<double,6,6>& C //compliance matrix
  //                                                   ) 
  Dune::FieldMatrix<double,6,6> general_anisotropic(const int& axis,
                                                    const double& angle,
                                                    Dune::FieldMatrix<double,6,6>& Compliance
                                                    ) 
  {

      std::string axis_string;
      switch (axis)
      {
        case 0: axis_string = "x - axis";
        case 1: axis_string = "y - axis";
        case 2: axis_string = "z - axis";
        default: break;
      }

      /**
       * @brief (optional) Tramsform compliance matrix to the correct frame.
       */
      if(abs(angle) > 1e-12){
        // std::cout << "material frame rotated around axis: " << axis << " by an angle of: " << angle << std::endl;
        std::cout << "material frame rotated around axis: " << axis_string << " by an angle of: " << angle << std::endl;
        Dune::FieldMatrix<double,6,6> C_rot = rotationMatrixCompliance(axis,angle);
        (Compliance.leftmultiply(C_rot)).rightmultiply(C_rot.transposed());
      }
      printmatrix(std::cout, Compliance, "Compliance matrix used:", "--");




      //--- return stiffness/elasticity tensor (in Voigt notation)
      Compliance.invert();
      // printmatrix(std::cout, Compliance, "Stiffness Tensor", "--");
      
      scaleStiffnessMatrix(Compliance);

      return Compliance;
  }
  //-------------------------------------------------------------------------------


  VoigtVector<double,3> applyElasticityTensor(const VoigtVector<double,3>& G, const int& phase)  const
  {
    VoigtVector<double,3> out(0);

    // printmatrix(std::cout, L_[0], "L_[0]", "--");
    L_[phase-1].mv(G,out);

    return out;
  }


  ////////////////////////////////////////////////////////////////////////////////////////
  MatrixRT prestrain(const int& phase,const Domain& x) const  
  {
    return prestrain_[phase-1](x);
  }


////////////////////////////////////////////////////////


  //--- Apply elasticityTensor_ to input Matrix G at material phase
 // input: Matrix G, material-phase
  MatrixRT ElasticityTensor(const MatrixRT& G, const int& phase) const      //deprecated
  {
    return elasticityTensor_(G,phase);   
  }

  int IndicatorFunction(const Domain& x) const
  {
    return indicatorFunction_(x);
  }


    ///////////////////////////////
    // Getter
    ///////////////////////////////
    Dune::ParameterTree getParameterSet() const {return parameterSet_;}
    // Func2Tensor getElasticityTensor() const {return elasticityTensor_;}
    // Func2TensorParam getElasticityTensor() const {return elasticityTensor_;}
    Func2TensorPhase getElasticityTensor() const {return elasticityTensor_;}
    MatrixPhase getPrestrainFunction() const {return prestrain_;} 

    auto getLocalIndicatorFunction() const {return localIndicatorFunction_;}   //get as localFunction
    auto getIndicatorFunctionGVF() const {return indicatorFunctionGVF_;}   //get as GridViewFunction
    auto getIndicatorFunction() const {return indicatorFunction_;}

    //getLameParameters()? .. Throw Exception if isotropic_ = false! 

  ///////////////////////////////
  // VTK - Writer
  ///////////////////////////////
  // -----------------------------------------------------------------
  // --- write material (grid)functions to VTK
  // void writeVTKMaterialFunctions(std::array<int, dim> nElements, const int level)
  void writeVTKMaterialFunctions(const int level)
  {
        std::string outputPath = parameterSet_.get("outputPath", "../../outputs");

        int subsamplingRefinement = parameterSet_.get<int>("subsamplingRefinement", 2);

        auto scalarP0FeBasis = makeBasis(gridView_,lagrange<0>());
        std::vector<double> indicatorFunction_CoeffP0;
        Dune::Functions::interpolate(scalarP0FeBasis, indicatorFunction_CoeffP0, indicatorFunction_);
        auto indicatorFunction_P0 = Dune::Functions::makeDiscreteGlobalBasisFunction<double>(scalarP0FeBasis, indicatorFunction_CoeffP0);
        
        Dune::SubsamplingVTKWriter<GridView> MaterialVtkWriter(gridView_, Dune::refinementLevels(subsamplingRefinement));
        
        MaterialVtkWriter.addCellData(
            indicatorFunction_P0,
            Dune::VTK::FieldInfo("materialIndicatorFunction_P0", Dune::VTK::FieldInfo::Type::scalar, 1));

        std::string baseName = parameterSet_.get("baseName", "CellProblem-result");
        MaterialVtkWriter.write(outputPath + "/" + baseName + "_MaterialFunction-level"+ std::to_string(level) );
        std::cout << "wrote data to file:" + outputPath + "/" + baseName + "_MaterialFunction-level"+ std::to_string(level) << std::endl;  
	      return;
  };

  // -----------------------------------------------------------------
  // --- write prestrain (grid)functions to VTK
  // void writeVTKPrestainFunctions(std::array<int, dim> nElements, const int level)
  // {
  //       std::string outputPath = parameterSet_.get("outputPath", "../../outputs");
  //       using VTKGridType = YaspGrid<dim, EquidistantOffsetCoordinates<double, dim> >;
  //       //VTKGridType grid_VTK({-1.0/2.0, -1.0/2.0, -1.0/2.0},{1.0/2.0, 1.0/2.0, 1.0/2.0},{80,80,80});
  //       VTKGridType grid_VTK({-1.0/2.0, -1.0/2.0, -1.0/2.0},{1.0/2.0, 1.0/2.0, 1.0/2.0},nElements);
        
  //       using GridViewVTK = VTKGridType::LeafGridView;
  //       const GridViewVTK gridView_VTK = grid_VTK.leafGridView();
        
  //       auto scalarP0FeBasis = makeBasis(gridView_VTK,lagrange<0>());
  //       auto scalarP1FeBasis = makeBasis(gridView_VTK,lagrange<1>());

  //       std::vector<double> prestrainFunction_CoeffP0;
  //       Functions::interpolate(scalarP0FeBasis, prestrainFunction_CoeffP0, prestrain_);
  //       auto prestrainFunction_P0 = Functions::makeDiscreteGlobalBasisFunction<double>(scalarP0FeBasis, prestrainFunction_CoeffP0);
        
  //       std::vector<double> prestrainFunction_CoeffP1;
  //       Functions::interpolate(scalarP1FeBasis, prestrainFunction_CoeffP1, prestrain_);
  //       auto prestrainFunction_P1 = Functions::makeDiscreteGlobalBasisFunction<double>(scalarP1FeBasis, prestrainFunction_CoeffP1);
        
  //       VTKWriter<GridView> MaterialVtkWriter(gridView_VTK);
        
  //       MaterialVtkWriter.addVertexData(
  //           prestrainFunction_P0,
  //           VTK::FieldInfo("prestrainFunction_P0", VTK::FieldInfo::Type::scalar, 1));    
  //       MaterialVtkWriter.addCellData(
  //           prestrainFunction_P1,
  //           VTK::FieldInfo("prestrainFunction_P1", VTK::FieldInfo::Type::scalar, 1));    

  //       MaterialVtkWriter.write(outputPath + "/PrestrainFunctions-level"+ std::to_string(level) );
  //       std::cout << "wrote data to file:" + outputPath + "/PrestrainFunctions-level" + std::to_string(level) << std::endl;  
	//       return;
  // };

  // void writeVTKPrestainFunctions(std::array<int, dim> nElements, const int level)
  // {
//     using VTKGridType = YaspGrid<dim, EquidistantOffsetCoordinates<double, dim> >;
// //     VTKGridType grid_VTK({-1.0/2.0, -1.0/2.0, -1.0/2.0},{1.0/2.0, 1.0/2.0, 1.0/2.0},{80,80,80});
// //     VTKGridType grid_VTK({-1.0/2.0, -1.0/2.0, -1.0/2.0},{1.0/2.0, 1.0/2.0, 1.0/2.0},{40,40,40});
//     VTKGridType grid_VTK({-1.0/2.0, -1.0/2.0, -1.0/2.0},{1.0/2.0, 1.0/2.0, 1.0/2.0},nElements);
//     using GridViewVTK = VTKGridType::LeafGridView;
//     const GridViewVTK gridView_VTK = grid_VTK.leafGridView();
   
//     FTKfillerContainer<dim> VTKFiller;
//     VTKFiller.vtkPrestrainNorm(gridView_VTK, B_Term, "PrestrainBNorm");
    
//     // WORKS Too 
//     VTKFiller.vtkProblemCell(gridView_VTK, B_Term, muLocal,"VTKProblemCell");;
    
    
//     // TEST 
//     auto scalarP0FeBasis = makeBasis(gridView_VTK,lagrange<0>());
//     auto scalarP1FeBasis = makeBasis(gridView_VTK,lagrange<1>());

//     std::vector<double> B_CoeffP0;
//     Functions::interpolate(scalarP0FeBasis, B_CoeffP0, B_Term);
//     auto B_DGBF_P0 = Functions::makeDiscreteGlobalBasisFunction<double>(scalarP0FeBasis, B_CoeffP0);
        

        
//     VTKWriter<GridView> PrestrainVtkWriter(gridView_VTK);
         
//     PrestrainVtkWriter.addCellData(
//             B_DGBF_P0,
//             VTK::FieldInfo("B_P0", VTK::FieldInfo::Type::scalar, 1));     
        
//     PrestrainVtkWriter.write(outputPath + "/PrestrainFunctions-level"+ std::to_string(level) );
//     std::cout << "wrote data to file:" + outputPath +"/PrestrainFunctions-level" + std::to_string(level) << std::endl; 

//    // };
  
  // MatrixRT ElasticityTensorGlobal(const MatrixRT& G, const Domain& x) const  //global Version (takes global coordinates)
  // {
  //   return MAT(G,x);
  // }

  // //--- apply elasticityTensor_ to input Matrix G at position x
  // MatrixRT applyElasticityTensor(const MatrixRT& G, const Domain& x) const  //global Version (takes global coordinates)
  // {
  //   // Python::Module module = Python::import("material");
  //   // auto indicatorFunctionTest = Python::make_function<double>(module.get("indicatorFunction"));
  //   // return elasticityTensor_(G,indicatorFunctionTest(x));
  //   // auto IFunction = localFunction(indicatorFunction_);
  //   // auto IFunction = this->getIndicatorFunction();
  //   // auto tmp = Dune::Functions::makeGridViewFunction(indicatorFunction_material_1, gridView_);
  //   // auto tmpLocal = LocalF
  //   // return elasticityTensor_(G,IFunction(x));
  //   // return elasticityTensor_(G,localIndicatorFunction_(x));
  //   return elasticityTensor_(G,indicatorFunction_(x));
  //   // return elasticityTensor_(G,indicatorFunctionGVF_(x));
  //   // int phase = indicatorFunction_(x);
  //   // auto tmp = elasticityTensor_(G,phase);
  //   // auto tmp = material_1(G,indicatorFunction_(x));
  //   // printmatrix(std::cout, material_1(G,indicatorFunction_(x)), "material_1(G,indicatorFunction_(x))", "--");
  //   // printmatrix(std::cout, elasticityTensor_(G,phase), "elasticityTensor_(G,phase)", "--");
  // }


//  template<class Domain>
// static MatrixRT MAT(const MatrixRT& G, const Domain& x)      //unfortunately not possible as a GridViewFunction? 
//   {
//       const FieldVector<double,3> mu = {80.0, 80.0, 60.0};
//       const FieldVector<double,3> lambda = {80.0, 80.0, 25.0};
//       double theta=0.25;
//       if (x[0] <(-0.5+theta) and x[2]<(-0.5+theta))
//          return 2.0 * mu[0] * sym(G) + lambda[0] * trace(sym(G)) * Id();
//       else if (x[1]<(-0.5+theta) and x[2]>(0.5-theta))
//          return 2.0 * mu[1] * sym(G) + lambda[1] * trace(sym(G)) * Id();
//       else 
//          return 2.0 * mu[2] * sym(G) + lambda[2] * trace(sym(G)) * Id();
//   }



}; // end class


#endif
